from typing import List, Union

from . import Photo, Video, VideoStream


class Media:
    r"""A Media object that represents a Weverse Media Post.

    It is not suggested to create a Media object manually, but rather through the
    following method: :class:`Weverse.objects.create_media_object`

    The information retrieved on Media is directly from the Weverse API and altered to fit this class.

    .. container:: operations

        .. describe:: x == y

            Checks if two Media objects have the same ID.

        .. describe:: x != y

            Checks if two Media objects do not have the same ID.

    Parameters
    ----------
    id: int
        ID of the Media post.
    community_id: int
        ID of the Community the media post was made in.
    body: str
        The media content AKA the body of the message.
    type: str
        The type of media post it is.
    thumbnail_path: str
        The (url??) of the thumbnail.
    title: str
        The title of the media post.
    level:
        The level of access the media post is categorized under.
    video_link: str
        The video link supplied under the media post.
    youtube_id: str
        The youtube video ID.


    Attributes
    -----------
    id: int
        ID of the Media post.
    community_id: int
        ID of the Community the media post was made in.
    body: str
        The media content AKA the body of the message.
    type: str
        The type of media post it is.
    thumbnail_path: str
        The (url??) of the thumbnail.
    title: str
        The title of the media post.
    level:
        The level of access the media post is categorized under.
    video_link: str
        The video link supplied under the media post.
    youtube_id: str
        The youtube video ID.
    photos: List[:ref:`Photo`]
        A list of photos under the media post.
    videos: List[:ref:`Video`]
        A list of videos under the media post.
    """
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.community_id = kwargs.get('communityId')
        self.body = kwargs.get('body')
        self.type = kwargs.get('type')
        self.thumbnail_path = kwargs.get('thumbnailPath')
        self.title = kwargs.get('title')
        self.level = kwargs.get('level')
        self.video_link = kwargs.get('extVideoPath')
        self.youtube_id = kwargs.get('youtubeId')
        self.photos: List[Photo] = kwargs.get('photo_objects') or []
        self.videos: List[Union[Video, VideoStream]] = kwargs.get('video_objects') or []

    def __eq__(self, other):
        """Check if the IDs of the Media objects are equal."""
        if not isinstance(other, Media):
            return NotImplemented

        return self.id == other.id

    def __ne__(self, other):
        """Check if the IDs of the Media objects are not equal."""
        return not self == other




from typing import List, Dict

from .models import Community, Artist, Tab, Notification, Post, Photo, Comment, Media, Video, Announcement, \
    VideoStream


def create_community_objects(current_communities: list, already_existing: Dict[int, Community] = None) -> dict:
    """Creates community objects based on a list of information sent in and returns the objects.

    :param current_communities: A list of communities from the endpoint being followed.
        Community information received from endpoint.
    :param already_existing: List[:ref:`Community`]
        Already existing Communities that should not be replaced.

    :returns: dict{community id: :ref:`Community`}
    """
    community_objects = {} if not already_existing else already_existing
    if current_communities:
        for community in current_communities:
            if community.get('id') in community_objects:
                continue

            kwargs = {
                'community_id': community.get('id'),
                'name': community.get('name'),
                'description': community.get('description'),
                'member_count': community.get('memberCount'),
                'home_banner': community.get('homeBannerImgPath'),
                'icon': community.get('iconImgPath'),
                'banner': community.get('bannerImgPath'),
                'full_name': community.get('fullname'),
                'fc_member': community.get('fcMember'),
                'show_member_count': community.get('showMemberCount')
            }
            community_object = Community(**kwargs)
            community_objects[community_object.id] = community_object
    return community_objects


def create_artist_objects(current_artists: list) -> list:
    """Creates artist objects based on a list of information sent in and returns the objects.

    :param current_artists: Artist information received from endpoint.
    :returns: List[:ref:`Artist`]
    """
    community_artists = []
    if current_artists:
        for artist in current_artists:
            kwargs = {
                'artist_id': artist.get('id'),
                'community_user_id': artist.get('communityUserId'),
                'name': artist.get('name'),
                'list_name': artist.get('listName'),
                'is_online': artist.get('isOnline'),
                'profile_nick_name': artist.get('profileNickName'),
                'profile_img_path': artist.get('profileImgPath'),
                'is_birthday': artist.get('isBirthday'),
                'group_name': artist.get('groupName'),
                'max_comment_count': artist.get('maxCommentCount'),
                'community_id': artist.get('communityId'),
                'is_enabled': artist.get('isEnabled'),
                'has_new_to_fans': artist.get('hasNewToFans'),
                'has_new_private_to_fans': artist.get('hasNewPrivateToFans'),
                'to_fan_last_id': artist.get('toFanLastId'),
                'to_fan_last_created_at': artist.get('toFanLastCreatedAt'),
                'to_fan_last_expire_in': artist.get('toFanLastExpireIn'),
                'birthday_img_url': artist.get('birthdayImgUrl')
            }
            community_artist = Artist(**kwargs)
            community_artists.append(community_artist)
    return community_artists


def create_tab_objects(current_tabs: list) -> list:
    """Creates tab objects based on a list of information sent in and returns the objects.

    :param current_tabs: Tab information received from endpoint.
    :returns: List[:ref:`Tab`]
    """
    community_tabs = []
    if current_tabs:
        for tab in current_tabs:
            kwargs = {
                'tab_id': tab.get('id'),
                'name': tab.get('name')
            }
            community_tab = Tab(**kwargs)
            community_tabs.append(community_tab)
    return community_tabs


def create_notification_objects(current_notifications: list) -> list:
    """Creates notification objects based on a list of information sent in and returns the objects.

    :param current_notifications: Notification information received from endpoint.
    :returns: List[:ref:`Notification`]
    """
    user_notifications = []
    if current_notifications:
        for notification in current_notifications:
            kwargs = {
                'notification_id': notification.get('id'),
                'message': notification.get('message'),
                'bold_element': notification.get('boldElement'),
                'community_id': notification.get('communityId'),
                'community_name': notification.get('CommunityName'),
                'contents_type': notification.get('contentsType'),
                'contents_id': notification.get('contentsId'),
                'notified_at': notification.get('notifiedAt'),
                'icon_image_url': notification.get('iconImageUrl'),
                'thumbnail_image_url': notification.get('thumbnailImageUrl'),
                'artist_id': notification.get('artistId'),
                'is_membership_content': notification.get('isMembershipContent'),
                'is_web_only': notification.get('isWebOnly'),
                'platform': notification.get('platform')
            }
            user_notification = Notification(**kwargs)
            user_notifications.append(user_notification)
    return user_notifications


def create_post_objects(current_posts: list, community: Community, new=False) -> list:
    """Creates post objects based on a list of posts sent in and the community and returns the objects.

    :param current_posts: Post information received from endpoint.
    :param community: :ref:`Community` that the post belongs in.
    :param new: :class:`bool` Whether or not the post is new.
    :returns: List[:ref:`Post`]
    """
    posts = []
    if current_posts:
        for post in current_posts:
            artist_comments = create_comment_objects(post.get('artistComments'))
            artist_photos = create_photo_objects(post.get('photos'))
            artist_videos = create_video_objects(post.get('attachedVideos'), community.id)
            artist_info = post.get('communityUser')
            community_artist_id = artist_info.get('id')
            kwargs = {
                'artist_id': artist_info.get('artistId'),
                'community_artist_id': community_artist_id,
                'post_id': post.get('id'),
                'community_tab_id': post.get('communityTabId'),
                'post_type': post.get('type'),
                'body': post.get('body'),
                'comment_count': post.get('commentCount'),
                'like_count': post.get('likeCount'),
                'max_comment_count': post.get('maxCommentCount'),
                'has_my_like': post.get('hasMyLike'),
                'has_my_bookmark': post.get('hasMyBookmark'),
                'created_at': post.get('createdAt'),
                'updated_at': post.get('updatedAt'),
                'is_locked': post.get('isLocked'),
                'is_blind': post.get('isBlind'),
                'is_active': post.get('isActive'),
                'is_private': post.get('isPrivate'),
                'photos': artist_photos,
                'videos': artist_videos,
                'is_hot_trending_post': post.get('isHotTrendingPost'),
                'is_limit_comment': post.get('isLimitComment'),
                'artist_comments': artist_comments
            }
            post_obj = Post(**kwargs)
            if new:
                posts.insert(0, post_obj)
            else:
                posts.append(post_obj)
            for comment in artist_comments:
                comment.post = post_obj
            for photo in artist_photos:
                photo.post = post_obj
            for video in artist_videos:
                video.post = post_obj
            for artist in community.artists:
                if artist.community_user_id == community_artist_id:
                    artist.posts.append(post_obj)
                    post_obj.artist = artist
    return posts


def create_video_objects(current_videos: list, community_id=None) -> list:
    """Creates & Returns video objects based on a list of videos.

    :param current_videos: Video information from api endpoint.
    :param community_id: Community ID
    :returns: List[:ref:`Video`]
    """
    videos = []
    if current_videos:
        for video in current_videos:
            kwargs = {
                'community_id': community_id,
                'video_url': video.get('videoUrl'),
                'thumbnail_url': video.get('thumbnailUrl'),
                'thumbnail_width': video.get('thumbnailWidth'),
                'thumbnail_height': video.get('thumbnailHeight'),
                'playtime': video.get('playTime'),
                'content_index': video.get('contentIndex'),
                'video_id': video.get('id'),
                'encoding_status': video.get('status'),
                'type': video.get('type'),
                'video_width': video.get("videoWidth"),
                'video_height': video.get('videoHeight'),
                'is_vertical': video.get("isVertical"),
                "caption_s3_paths": video.get("captionS3Paths"),
                "level": video.get("community"),
                "hls_path": video.get("hlsPath"),
                "dash_path": video.get("dashPath")
            }

            if kwargs["hls_path"] is not None:
                video_obj = VideoStream(**kwargs)
            else:
                video_obj = Video(**kwargs)
            videos.append(video_obj)
    return videos


def create_photo_objects(current_photos: list) -> list:
    """Creates & Returns photo objects based on a list of photos

    :param current_photos: photo information from endpoint.
    :returns: List[:ref:`Photo`]
    """
    photos = []
    if current_photos:
        for photo in current_photos:
            kwargs = {
                'photo_id': photo.get('id'),
                'media_id': photo.get('mediaId'),
                'content_index': photo.get('contentIndex'),
                'thumbnail_img_url': photo.get('thumbnailImgUrl'),
                'thumbnail_img_width': photo.get('thumbnailImgWidth'),
                'thumbnail_img_height': photo.get('thumbnailImgHeight'),
                'original_img_url': photo.get('orgImgUrl'),
                'original_img_width': photo.get('orgImgWidth'),
                'original_img_height': photo.get('orgImgHeight'),
                'file_name': photo.get('downloadImgFilename')
            }
            photo_obj = Photo(**kwargs)
            photos.append(photo_obj)
    return photos


def create_comment_objects(current_comments: list) -> list:
    """Creates & Returns comment objects based on a list of comments

    :param current_comments: comment information from endpoint.
    :returns: List[:ref:`Comment`]
    """
    comments = []
    if current_comments:
        for comment in current_comments:
            kwargs = {
                'comment_id': comment.get('id'),
                'body': comment.get('body'),
                'comment_count': comment.get('commentCount'),
                'like_count': comment.get('likeCount'),
                'has_my_like': comment.get('hasMyLike'),
                'is_blind': comment.get('isBlind'),
                'post_id': comment.get('postId'),
                'created_at': comment.get('createdAt'),
                'updated_at': comment.get('updatedAt')
            }
            comment_obj = Comment(**kwargs)
            comments.append(comment_obj)

    return comments


def create_announcement_object(announcement_info: dict) -> Announcement:
    """Creates and returns an announcement object

    :param announcement_info: Announcement information from endpoint.
    :returns: :ref:`Announcement`
    """
    return Announcement(**announcement_info)


def create_media_object(media_info: dict, ignore_photos=False, ignore_videos=False) -> Media:
    """Creates and returns a media object

    :param media_info: media information from endpoint.
    :param ignore_photos: Whether to ignore the photos that belong in the media object. (Other methods can
        create it themselves.)
    :param ignore_videos: Whether to ignore the videos that belong in the media object. (Other methods can
        create it themselves.)
    :returns: :ref:`Media`
    """
    if media_info.get("type") == "PHOTO" and not ignore_photos:
        photos = media_info.get("photos")

        if photos:
            media_info["photo_objects"] = create_photo_objects(photos)

    if media_info.get("type") == "VIDEO" and not ignore_videos:
        video = media_info.get("video")

        if video:
            media_info["video_objects"] = create_video_objects([video], media_info.get("communityId"))

    return Media(**media_info)


def iterate_community_media_categories(all_media_categories: dict) -> [List[Media], List[dict]]:
    """Iterates through community media categories, creates Media posts and returns a list of them.

    :param all_media_categories: A dict containing media posts that are filtered by category.
    :returns: [List[:ref:`Media`], List[:ref:`dict`]] A list of Video Media objects and a list of dicts containing
        photo media objects to later make own calls on to retrieve photos.
    """
    photo_media_dicts = []
    video_media_objects = []
    media_category = all_media_categories.get("mediasByCategory")
    for media_dicts in media_category:
        medias_list = media_dicts.get("medias")
        # media_category_info = media_dicts.get("mediaCategory")
        for media in medias_list:
            # The information we are receiving from the endpoint does not give us any information about the photos,
            # therefore, we need our caller method to take care of loading the media photos for us.
            if media.get("type") == "PHOTO":
                photo_media_dicts.append(media)
            else:
                video_media_objects.append(create_media_object(media))
    return [video_media_objects, photo_media_dicts]



from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from Weverse.models import Community


class Artist:
    r"""An Artist object that represents a Weverse Artist that belongs in a community.

    It is not suggested to create an Artist manually, but rather through the
    following method: :class:`Weverse.objects.create_artist_objects`

    The information retrieved on an Artist is directly from the Weverse API and altered to fit this class.

    .. container:: operations

        .. describe:: x == y

            Checks if two Artists have the same ID.

        .. describe:: x != y

            Checks if two Artists do not have the same ID.

        .. describe:: str(x)

            Returns the Artist's primary name.

    Parameters
    ----------
    id: :class:`int`
        The Artist ID.
    community_user_id: :class:`int`
        Artist's ID in the community.
    name: :class:`str`
        The Primary Artist Name.
    list_name: :class:`list`
        A list of names for the Artist.
    is_online: :class:`bool`
        Whether the Artist is currently online
    profile_nick_name: :class:`str`
        Artist nickname.
    profile_img_path: :class:`str`
        Image URL for the Artist's profile.
    is_birthday: :class:`bool`
        Whether it is the Artist's birthday.
    group_name: :class:`str`
        The group name the Artist is associated with.
    max_comment_count: :class:`int`
        The maximum amount of comments this Artist can post.
    community_id: :class:`int`
        The ID of the community this Artist object was selected from.
    is_enabled: :class:`bool`
        If the Artist account is enabled.
    has_new_to_fans: :class:`bool`
        If the Artist has a new post for fans.
    has_new_private_to_fans: :class:`bool`
        If the Artist has a new private post for fans.
    to_fan_last_id: :class:`int`
        The latest tofan post ID.
    to_fan_last_created_at:
        When the artist's last tofan post was created.
    to_fan_last_expire_in:
        When the artist's last tofan post expires.
    birthday_img_url: :class:`str`
        A direct image url to the artist's birthday image.
    community: Community
        The community the Artist is in.
    posts: :class:`list`
        A list of posts the Artist has.

    Attributes
    -----------
    id: int
        The Artist ID.
    community_user_id: int
        Artist's ID in the community.
    name: str
        The Primary Artist Name.
    list_name: list
        A list of names for the Artist.
    is_online: bool
        Whether the Artist is currently online
    profile_nick_name: str
        Artist nickname.
    profile_img_path: str
        Image URL for the Artist's profile.
    is_birthday: bool
        Whether it is the Artist's birthday.
    group_name: str
        The group name the Artist is associated with.
    max_comment_count: int
        The maximum amount of comments this Artist can post.
    community_id: int
        The ID of the community this Artist object was selected from.
    is_enabled: bool
        If the Artist account is enabled.
    has_new_to_fans: bool
        If the Artist has a new post for fans.
    has_new_private_to_fans: bool
        If the Artist has a new private post for fans.
    to_fan_last_id: int
        The latest tofan post ID.
    to_fan_last_created_at:
        When the artist's last tofan post was created.
    to_fan_last_expire_in:
        When the artist's last tofan post expires.
    birthday_img_url: str
        A direct image url to the artist's birthday image.
    community: Community
        The community the Artist is in.
    posts: list
        A list of posts the Artist has.
    """
    def __init__(self, **kwargs):
        self.id = kwargs.get('artist_id')
        self.community_user_id = kwargs.get('community_user_id')
        self.name = kwargs.get('name')
        self.list_name = kwargs.get('list_name')
        self.is_online = kwargs.get('is_online')
        self.profile_nick_name = kwargs.get('profile_nick_name')
        self.profile_img_path = kwargs.get('profile_img_path')
        self.is_birthday = kwargs.get('is_birthday')
        self.group_name = kwargs.get('group_name')
        self.max_comment_count = kwargs.get('max_comment_count')
        self.community_id = kwargs.get('community_id')
        self.is_enabled = kwargs.get('is_enabled')
        self.has_new_to_fans = kwargs.get('has_new_to_fans')
        self.has_new_private_to_fans = kwargs.get('has_new_private_to_fans')
        self.to_fan_last_id = kwargs.get('to_fan_last_id')
        self.to_fan_last_created_at = kwargs.get('to_fan_last_created_at')
        self.to_fan_last_expire_in = kwargs.get('to_fan_last_expire_in')
        self.birthday_img_url = kwargs.get('birthday_img_url')
        self.community: Optional[Community] = None
        self.posts = []

    def __eq__(self, other):
        """Check if the IDs of the Artists are equal."""
        if not isinstance(other, Artist):
            return NotImplemented

        return self.id == other.id

    def __ne__(self, other):
        """Check if the IDs of the Artists are not equal."""
        return not self == other

    def __str__(self):
        """Returns the Primary Artist Name."""
        return f"{self.name}"



class InvalidToken(Exception):
    """An Exception Raised When an Invalid Token was Supplied."""
    def __init__(self):
        super(InvalidToken, self).__init__("An Invalid Bearer Token was Supplied to Weverse.")


class PageNotFound(Exception):
    r"""
    An Exception Raised When a link was not found.

    Parameters
    ----------
    url: :class:`str`
        The link that was not found.
    """
    def __init__(self, url):
        super(PageNotFound, self).__init__(url + "was an invalid link.")


class BeingRateLimited(Exception):
    """An Exception Raised When Weverse Is Being Rate-Limited."""
    def __init__(self):
        super(BeingRateLimited, self).__init__("Weverse is rate-limiting the requests.")


class LoginFailed(Exception):
    """An Exception raised when the login failed."""
    def __init__(self, msg: str = "The login process for Weverse had failed."):
        super(LoginFailed, self).__init__(msg)


class InvalidCredentials(Exception):
    """An Exception raised when no valid credentials were supplied."""
    def __init__(self, msg: str = "The credentials for a token or a username/password could not be found."):
        super(InvalidCredentials, self).__init__(msg)


class NoHookFound(Exception):
    """An Exception raised when a loop for the hook was started but did not actually have a hook method."""
    def __init__(self, msg: str = "No Hook was passed into the Weverse client."):
        super(NoHookFound, self).__init__(msg)



from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from Weverse.models import Artist, Tab


class Community:
    r"""A Community object that represents a Weverse Community.

    It is not suggested to create a Community manually, but rather through the
    following method: :class:`Weverse.objects.create_community_objects`

    The information retrieved on a Community is directly from the Weverse API and altered to fit this class.

    .. container:: operations

        .. describe:: x == y

            Checks if two Communities have the same ID.

        .. describe:: x != y

            Checks if two Communities do not have the same ID.

        .. describe:: str(x)

            Returns the Community's name.

    Parameters
    ----------
    id: :class:`int`
        The Community ID.
    name: :class:`str`
        The Community Name.
    description: :class:`str`
        Description of the Community.
    member_count: :class:`int`
        Amount of members in the community.
    home_banner: :class:`str`
        Direct Image URL to the home banner.
    icon: :class:`str`
        Direct Image URL to the Icon.
    Banner: :class:`str`
        Direct Image URL to the Banner.
    full_name: :class:`str`
        Full Name of the Community.
    fc_member: bool
        If a special membership is required to join.
    show_member_count: bool
        If the member count is visible.

    Attributes
    -----------
    id: :class:`int`
        The Community ID.
    name: :class:`str`
        The Community Name.
    description: :class:`str`
        Description of the Community.
    member_count: :class:`int`
        Amount of members in the community.
    home_banner: :class:`str`
        Direct Image URL to the home banner.
    icon: :class:`str`
        Direct Image URL to the Icon.
    Banner: :class:`str`
        Direct Image URL to the Banner.
    full_name: :class:`str`
        Full Name of the Community.
    fc_member: bool
        If a special membership is required to join.
    show_member_count: bool
        If the member count is visible.
    artists: List[Artist]
        List of artists the community has.
    tabs: List[Tab]
        The Tabs the community has.
    """
    def __init__(self, **kwargs):
        self.id = kwargs.get('community_id')
        self.name = kwargs.get('name')
        self.description = kwargs.get('description')
        self.member_count = kwargs.get('member_count')
        self.home_banner = kwargs.get('home_banner')
        self.icon = kwargs.get('icon')
        self.banner = kwargs.get('banner')
        self.full_name = kwargs.get('full_name')
        self.fc_member = kwargs.get('fc_member')
        self.show_member_count = kwargs.get('show_member_count')
        self.artists: List[Artist] = []
        self.tabs: List[Tab] = []

    def __eq__(self, other):
        """Check if the IDs of the Communities are equal."""
        if not isinstance(other, Community):
            return NotImplemented

        return self.id == other.id

    def __ne__(self, other):
        """Check if the IDs of the Communities are not equal."""
        return not self == other

    def __str__(self):
        """Returns the Community Name"""
        return f"{self.name}"



from typing import Optional, List

from . import Video, Photo, Artist, Comment


class Post:
    r"""A Post object that represents a Weverse Post.

    It is not suggested to create a Post manually, but rather through the
    following method: :class:`Weverse.objects.create_post_objects`

    The information retrieved on a Post is directly from the Weverse API and altered to fit this class.

    .. container:: operations

        .. describe:: x == y

            Checks if two Post objects have the same ID.

        .. describe:: x != y

            Checks if two Post objects do not have the same ID.

        .. describe:: str(x)

            Returns the Post body message.

        .. describe:: len(x)

            Returns the amount of images (not videos) available.

    Parameters
    ----------
    id: int
        The ID of the post.
    community_tab_id: int
        The tab the post is under.
    type: str
        The type of Post.
    body: str
        Body Message on the Post.
    comment_count: int
        Current amount of comments on the Post
    like_count: int
        Current amount of likes on the Post
    max_comment_count: int
        Maximum amount of comments that can be on the Post
    has_my_like: bool
        If the client user has the post liked.
    has_my_bookmark: bool
        If the client user has the post bookmarked.
    created_at:
        When the post was created
    updated_at:
        When the post was last modified.
    is_locked: bool
        Whether the post is locked.
    is_blind: bool
        Whether the post is visible?? Unknown
    is_active: bool
        Whether the post is active.
    is_private: bool
        Whether the post is private.
    photos: List[:ref:`Photo`]
        A list of photos under the post.
    videos: List[:ref:`Video`]
        A list of videos under the post.
    is_hot_trending_post: bool
        If the post is trending.
    is_limit_comment: bool
        If the comments are limited.
    artist_comments: List[:ref:`Comment`]
        The Artist comments under the post.
    community_artist_id: int
        The Community Artist ID that made the post.
    artist_id: int
        The ID of the Artist that made the post.

    Attributes
    -----------
    id: int
        The ID of the post.
    community_tab_id: int
        The tab the post is under.
    type: str
        The type of Post.
    body: str
        Body Message on the Post.
    comment_count: int
        Current amount of comments on the Post
    like_count: int
        Current amount of likes on the Post
    max_comment_count: int
        Maximum amount of comments that can be on the Post
    has_my_like: bool
        If the client user has the post liked.
    has_my_bookmark: bool
        If the client user has the post bookmarked.
    created_at:
        When the post was created
    updated_at:
        When the post was last modified.
    is_locked: bool
        Whether the post is locked.
    is_blind: bool
        Whether the post is visible?? Unknown
    is_active: bool
        Whether the post is active.
    is_private: bool
        Whether the post is private.
    photos: List[:ref:`Photo`]
        A list of photos under the post.
    videos: List[:ref:`Video`]
        A list of videos under the post.
    is_hot_trending_post: bool
        If the post is trending.
    is_limit_comment: bool
        If the comments are limited.
    artist_comments: List[:ref:`Comment`]
        The Artist comments under the post.
    community_artist_id: int
        The Community Artist ID that made the post.
    artist_id: int
        The ID of the Artist that made the post.
    artist: Artist
        The Artist Object the post belongs to.
    """
    def __init__(self, **kwargs):
        self.id = kwargs.get('post_id')
        self.community_tab_id = kwargs.get('community_tab_id')
        self.type = kwargs.get('post_type')
        self.body = kwargs.get('body')
        self.comment_count = kwargs.get('comment_count')
        self.like_count = kwargs.get('like_count')
        self.max_comment_count = kwargs.get('max_comment_count')
        self.has_my_like = kwargs.get('has_my_like')
        self.has_my_bookmark = kwargs.get('has_my_bookmark')
        self.created_at = kwargs.get('created_at')
        self.updated_at = kwargs.get('updated_at')
        self.is_locked = kwargs.get('is_locked')
        self.is_blind = kwargs.get('is_blind')
        self.is_active = kwargs.get('is_active')
        self.is_private = kwargs.get('is_private')
        self.photos: List[Photo] = kwargs.get('photos')
        self.videos: List[Video] = kwargs.get('videos')
        self.is_hot_trending_post = kwargs.get('is_hot_trending_post')
        self.is_limit_comment = kwargs.get('is_limit_comment')
        self.artist_comments: Optional[List[Comment]] = kwargs.get('artist_comments')
        self.community_artist_id = kwargs.get('community_artist_id')
        self.artist_id = kwargs.get('artist_id')
        self.artist: Optional[Artist] = None

    def __eq__(self, other):
        """Check if the IDs of the Post objects are equal."""
        if not isinstance(other, Post):
            return NotImplemented

        return self.id == other.id

    def __ne__(self, other):
        """Check if the IDs of the Post objects are not equal."""
        return not self == other

    def __str__(self):
        """Returns the Post body message."""
        return f"{self.body}"

    def __len__(self):
        """Returns the amount of images (not videos) available."""
        return len(self.photos) or len(self.videos) or len(self.body)



import asyncio
from typing import List, Optional, Union, Dict

from . import create_artist_objects, create_tab_objects, InvalidToken, LoginFailed
from .models import Artist as w_Artist, \
    Comment as w_Comment, Media as w_Media, Notification as w_Notification, Photo as w_Photo, Post as w_Post, \
    Tab as w_Tab, Community as w_Community, Video as w_Video, Announcement as w_Announcement

from base64 import b64decode, b64encode
from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_OAEP


class WeverseClient:
    """
    Abstract & Parent Client for connecting to Weverse and creating the internal cache.

    Do not create an object directly from this class.
    Instead, create a :class:`Weverse.WeverseClientSync` or :class:`Weverse.WeverseClientAsync`
    object since those are concrete.

    Parameters
    ----------
    verbose: bool
        Whether to print out verbose messages.
    web_session:
        An aiohttp or requests client session.
    authorization: :class:`str`
        The account token to connect to the Weverse API. In order to find your token,
        please refer to :ref:`account_token`
    username: :class:`str`
        The email or username associated with the account.
    password: :class:`str`
        The password associated with the account.
    hook:
        A passed in method that will be called every time there is a new notification.
        This method must take in a list of :class:`models.Notification` objects.

    Attributes
    -----------
    verbose: bool
        Whether to print out verbose messages.
    web_session:
        An aiohttp or requests client session.
    user_notifications: list
        Most recent notifications of the account connected.
    cache_loaded: bool
        Whether the Internal Weverse Cache is fully loaded.
        This will change for a split moment when grabbing a new post.
    user_endpoint: str
        User info endpoint.
    all_posts: dict(Post)
        All posts in cache where the Post ID is the key and the value is the Post Object
    all_artists: dict(Artist)
        All artists in cache where the Artist ID is the key and the value is the Artist Object
    all_comments: dict(Comment)
        All comments in cache where the Comment ID is the key and the value is the Comment Object
    all_notifications: dict(Notification)
        All notifications in cache where the Notification ID is the key and the value is the Notification Object
    all_photos: dict(Photo)
        All photos in cache where the Photo ID is the key and the value is the Photo Object
    all_communities: dict(Community)
        All communities in cache where the Community ID is the key and the value is the Community Object
    all_media: dict(Media)
        All media in cache where the Media ID is the key and the value is the Media Object
    all_tabs: dict(Tab)
         All tabs in cache where the Tab ID is the key and the value is the Tab Object
    all_videos: dict(Video)
        All videos in cache where the Video URL is the key and the value is the Video Object
    all_announcements: dict(Announcement)
        All announcements/notices in cache where the Announcement ID is the key and the value is the Announcement Object
   """
    def __init__(self, **kwargs):
        self.verbose = kwargs.get('verbose')
        self.web_session = kwargs.get('web_session')
        self.__token = kwargs.get('authorization')
        self.user_notifications = []

        self._old_notifications = []
        self._headers = self.__get_headers()

        self.__login_payload = {
            "grant_type": "password",
            "client_id": "weverse-test",
            "username": kwargs.get("username"),
            "password": self.__get_encrypted_password(kwargs.get("password"))
        }

        self._refresh_payload = {
            "grant_type": "refresh_token",
            "client_id": "weverse-test",
            "refresh_token": None
        }

        self._login_url = "https://accountapi.weverse.io/api/v1/oauth/token"
        self._api_url = "https://weversewebapi.weverse.io/wapi/v1/"
        self._api_communities_url = self._api_url + "communities/"  # endpoint for communities
        self._api_notifications_url = self._api_url + "stream/notifications/"  # endpoint for user notifications
        # endpoint for checking new user notifications
        self._api_new_notifications_url = self._api_notifications_url + "has-new/"
        self._api_all_artist_posts_url = "posts/artistTab/"  # Artist Feed from a community
        self._api_stream_url = self._api_url + "/stream/community/"  # followed by community id and mediaTab.
        self._api_media_tab = "mediaTab/categorical?countByCategory=100000"
        self._api_artist_to_fans = "posts/tofans/"
        # endpoint for information about ALL communities and ALL idols.
        self._api_all_communities_info_url = self._api_communities_url + "info/"
        self.cache_loaded = False
        self._user_endpoint = "https://weversewebapi.weverse.io/wapi/v1/users/me"

        self.all_posts: Dict[int, w_Post] = {}
        self.all_artists: Dict[int, w_Artist] = {}
        self.all_comments: Dict[int, w_Comment] = {}
        self.all_notifications: Dict[int, w_Notification] = {}
        self.all_photos: Dict[int, w_Photo] = {}
        self.all_communities: Dict[int, w_Community] = {}
        self.all_media: Dict[int, w_Media] = {}
        self.all_tabs: Dict[int, w_Tab] = {}
        # Videos have the url as the key due to no unique ID.
        self.all_videos: Dict[str, w_Video] = {}
        self.all_announcements: Dict[int, w_Announcement] = {}

        self._hook = kwargs.get("hook")
        self._hook_loop = False
        self._expired_token = False
        self.__exception_to_raise = None

    @property
    def _login_info_exists(self) -> bool:
        """Whether login info is present."""
        return self.__login_payload["username"] and self.__login_payload["password"]

    @property
    def _token_exists(self) -> bool:
        """Whether a token is present."""
        return bool(self.__token)

    @property
    def _refresh_token_exists(self) -> bool:
        """Whether a refresh token is present."""
        return bool(self._refresh_payload["refresh_token"])

    @property
    def public_weverse_key(self) -> str:
        """Hard-coded weverse key"""
        return "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAu/OhimOynajYomJmBsNvQxSDwekunsp986l7s/zMN/8jHXFlTqT79ZOsOwzVdZcKnkWYXwJg4nhIFpaIsPzklQCImp2kfKUJQV3jzw7/Qtq6NrOOh9YBADr+b99SHYcc7E7cDHjGXgWlC5jEI9h80R822wBU0HcbODkAQ3uosvFhSq3gLpxwdimesZofkJ5ZbAmGIMj1GEWAfMGA49mxkv/cDFWry+6FM4mUW6A0301QUg4wK/8n6RrzRj1NUkevZj1smizHeqmBE+0BU5H/fR9HclErx3LMHlVlxSgEEEjNUx3B0bLO0OHppmEb4B3Tk1O3ZsquYyqZyb2lBTbrQwIDAQAB"

    def stop(self):
        """Stop the hook loop."""
        self._hook_loop = False

    def _login(self, method):
        """
        Requests a login.
        Will not return response.
        The method should handle the response itself.
        Parameters
        ----------
        method:
            The async/sync method to call. Should be able to take in the login payload.
        """
        if not asyncio.iscoroutinefunction(method):
            method(self.__login_payload)
        else:
            asyncio.create_task(method(self.__login_payload))

    def _set_exception(self, exception: Exception):
        """
        Set an exception from a task that occurred.
        Parameters
        ----------
        exception: The Exception that was raised.
        """
        self.__exception_to_raise = exception

    async def _wait_for_login(self, timeout=15):
        """
        Will wait until the client is logged in or the timeout timer is exceed.
        Parameters
        ----------
        timeout: int
            Amount of seconds before an exception is raised.
        :raises: :class:`Weverse.error.LoginFailed` Login process had failed.
        :raises: :class:`asyncio.exceptions.TimeoutError` Waited too long for a login.
        """
        seconds_passed = 0
        while not self._refresh_token_exists or self._expired_token:
            if self.__exception_to_raise:
                if isinstance(self.__exception_to_raise, (LoginFailed, asyncio.exceptions.TimeoutError)):
                    exception = self.__exception_to_raise
                    self.__exception_to_raise = None
                    # if an exception was raised from here, the actual exception occurred in a task.
                    raise exception
            await asyncio.sleep(1)
            seconds_passed += 1

            if seconds_passed > timeout:
                self._set_exception(asyncio.exceptions.TimeoutError())

    def get_new_notifications(self) -> List[w_Notification]:
        """Will get the new notifications from the last notification check.

        Should only be used after :ref:`check_new_user_notifications` OR :ref:`update_cache_from_notification`.

        :returns: List[:ref:`Notification`]
        """
        return [noti for noti in self.user_notifications if noti not in self._old_notifications]

    def check_status(self, status, url) -> bool:
        """
        Confirm the status of a URL

        :param status: Status code of url connection
        :param url: Link that we connected to.
        :return: True if the connection was a success.
        :raises: :ref:`invalid_token_exc` if there was an invalid token.
        """
        if status == 200:
            return True
        elif status == 401:
            self._expired_token = True
            # raise InvalidToken
        elif status == 404:
            if self.verbose:
                # raise error.PageNotFound
                print("WARNING (NOT CRITICAL): " + url + " was not found.")
        else:
            if self.verbose:
                print("WARNING (NOT CRITICAL): " + url + " Failed to load. [Status: " + str(status) + "]")

    @staticmethod
    def process_community_artists_and_tabs(community, response_text_as_dict):
        """
        Process the community artists and tabs and add them to their respective communities.

        :param community: Community object
        :param response_text_as_dict: Response text of connection to endpoint but as a dict.
        """
        # split up response dict by artists and tabs
        community_artists = response_text_as_dict.get('artists')
        community_tabs = response_text_as_dict.get('tabs')

        # create artist and tab objects under the community
        community.artists = create_artist_objects(community_artists)
        community.tabs = create_tab_objects(community_tabs)

        # set the community for each artist
        for artist in community.artists:
            artist.community = community

    def get_artist_by_id(self, artist_id) -> Optional[w_Artist]:
        """
        Get artist by their ID.

        :param artist_id: The artist's ID
        :returns: Optional[:ref:`Artist`]
        """
        artist = self.all_artists.get(artist_id)
        if not artist:
            for t_artist in self.all_artists.values():
                if t_artist.community_user_id == artist_id:
                    artist = t_artist
        return artist

    def get_tab_by_id(self, tab_id) -> Optional[w_Tab]:
        """
        Get tab by their ID.

        :param tab_id: The tab ID
        :returns: Optional[:ref:`Tab`]
        """
        return self.all_tabs.get(tab_id)

    def get_post_by_id(self, post_id) -> Optional[w_Post]:
        """
        Get a post by the ID

        :param post_id: Post ID
        :returns: Optional[:ref:`Post`]
        """
        return self.all_posts.get(post_id)

    def get_comment_by_id(self, comment_id) -> Optional[w_Comment]:
        """
        Get a comment by the ID
        :param comment_id: Comment ID
        :returns: Optional[:ref:`Comment`]
        """
        return self.all_comments.get(comment_id)

    def get_notification_by_id(self, notification_id) -> Optional[w_Notification]:
        """
        Get a notification by the ID

        :param notification_id: Notification ID
        :returns: Optional[:ref:`Notification`]
        """
        return self.all_notifications.get(notification_id)

    def get_photo_by_id(self, photo_id) -> Optional[w_Photo]:
        """
        Get a photo by the ID

        :param photo_id: Photo ID
        :returns: Optional[:ref:`Photo`]
        """
        return self.all_photos.get(photo_id)

    def get_video_by_url(self, video_url) -> Optional[w_Video]:
        """
        Get a video by the direct URL.

        :param video_url: URL of the video
        :return: Optional[:ref:`Video`]
        """
        return self.all_videos.get(video_url)

    def get_community_by_id(self, community_id) -> Optional[w_Community]:
        """
        Get a community by the ID.

        :param community_id: Community ID
        :returns: Optional[:ref:`Community`]
        """
        return self.all_communities.get(community_id)

    def get_media_by_id(self, media_id) -> Optional[w_Media]:
        """
        Get Media by the ID

        :param media_id: Media ID
        :returns: Optional[:ref:`Media`]
        """
        return self.all_media.get(media_id)

    def get_announcement_by_id(self, announcement_id) -> Optional[w_Announcement]:
        """
        Get Announcement by the ID

        :param announcement_id: Media ID
        :returns: Optional[:ref:`Announcement`]
        """
        return self.all_announcements.get(announcement_id)

    def _add_media_to_cache(self, media_objects: List[w_Media]):
        """
        Will add media objects and it's photos to cache.

        :param media_objects:
        """
        for media in media_objects:
            self.all_media[media.id] = media
            if media.photos:
                for photo in media.photos:
                    self.all_photos[photo.id] = photo

    def __get_encrypted_password(self, password):
        """
        Get the encrypted password from Weverse's public key.

        :param password: The password to encrypt with Weverse's public key.
        :return: Encrypted password.
        """
        if not password:
            return

        key_der = b64decode(self.public_weverse_key)
        pub_key = RSA.importKey(key_der)
        cipher = PKCS1_OAEP.new(pub_key)
        cipher_text = cipher.encrypt(password.encode())
        encrypted_pass = b64encode(cipher_text)
        password = encrypted_pass.decode("utf-8")
        return password

    def _set_token(self, token):
        """
        Set the token used for endpoints.
        Parameters
        ----------
        token: str
            New token used for endpoints
        """
        self.__token = token
        self._headers = self.__get_headers()  # update headers

    def __get_headers(self):
        return {'Authorization': f"Bearer {self.__token}"}

    def _set_refresh_token(self, refresh_token: str):
        """
        Set the refresh token.
        Parameters
        ----------
        refresh_token: str
            The refresh token to pass when logging in.
        """
        self._refresh_payload["refresh_token"] = refresh_token

    @staticmethod
    def determine_notification_type(notification: Union[w_Notification, str]) -> str:
        """
        Determine the post type based on the notification body or the Notification object itself.

        Since notifications do not differentiate between Posts and Comments, this is for that purpose.

        :param notification: The message body of the notification or the notification itself.
        :returns: A string with either
            "comment", "media", "post", or "announcement".
        """
        if isinstance(notification, w_Notification):
            notification_body = notification.message
        else:
            notification_body = notification

        comment_body_triggers = ["commented on", "replied to", "포스트에 댓글을 작성했습니다", "답글을 작성했습니다."]
        post_body_triggers = ["님이 포스트를 작성했습니다", "created a new post!", "shared a moment with you",
                              "모먼트가 도착했습니다"]
        media_triggers = ["Check out the new media", "새로운 미디어"]

        announcement_body_triggers = ["New announcement", "NOTICE:", "(광고)", "(AD)"]

        for trigger in comment_body_triggers:
            if trigger in notification_body:
                return "comment"

        for trigger in post_body_triggers:
            if trigger in notification_body:
                return "post"

        for trigger in media_triggers:
            if trigger in notification_body:
                return "media"

        for trigger in announcement_body_triggers:
            if trigger in notification_body:
                return "announcement"



class Notification:
    r"""A Media object that represents a Weverse Media Post.

    It is not suggested to create a Notification object manually, but rather through the
    following method: :class:`Weverse.objects.create_notification_objects`

    The information retrieved on a Notification is directly from the Weverse API and altered to fit this class.

    .. container:: operations

        .. describe:: x == y

            Checks if two Notifications have the same ID.

        .. describe:: x != y

            Checks if two Notifications do not have the same ID.

    Parameters
    ----------
    id: int
        The id of the notification.
    message: str
        The message of the notification.
    bold_element: str
        The bolded element in the notification.
    community_id: int
        The community id associated with the notification.
    community_name: str
        The community name associated with the notification.
    contents_type: str
        The type of post it is.
    contents_id: int
        The id of the content post.
    notified_at:
        The time the notification was triggered.
    icon_image_url: str
        Icon image url of the notification.
    thumbnail_image_url: str
        Thumbnail url of the notification.
    artist_id: int
        The ID of the Artist that released the content.
    is_membership_content: bool
        If the content is exclusive to members.
    is_web_only: bool
        Whether the notification is only available directly on the website.
    platform: str
        The platform of the notification.


    Attributes
    -----------
    id: int
        The id of the notification.
    message: str
        The message of the notification.
    bold_element: str
        The bolded element in the notification.
    community_id: int
        The community id associated with the notification.
    community_name: str
        The community name associated with the notification.
    contents_type: str
        The type of post it is.
    contents_id: int
        The id of the content post.
    notified_at:
        The time the notification was triggered.
    icon_image_url: str
        Icon image url of the notification.
    thumbnail_image_url: str
        Thumbnail url of the notification.
    artist_id: int
        The ID of the Artist that released the content.
    is_membership_content: bool
        If the content is exclusive to members.
    is_web_only: bool
        Whether the notification is only available directly on the website.
    platform: str
        The platform of the notification.
    """
    def __init__(self, **kwargs):
        self.id = kwargs.get('notification_id')
        self.message = kwargs.get('message')
        self.bold_element = kwargs.get('bold_element')
        self.community_id = kwargs.get('community_id')
        self.community_name = kwargs.get('community_name')
        self.contents_type = kwargs.get('contents_type')
        self.contents_id = kwargs.get('contents_id')
        self.notified_at = kwargs.get('notified_at')
        self.icon_image_url = kwargs.get('icon_image_url')
        self.thumbnail_image_url = kwargs.get('thumbnail_image_url')
        self.artist_id = kwargs.get('artist_id')
        self.is_membership_content = kwargs.get('is_membership_content')
        self.is_web_only = kwargs.get('is_web_only')
        self.platform = kwargs.get('platform')

    def __eq__(self, other):
        """Check if the IDs of the Notifications are equal."""
        if not isinstance(other, Notification):
            return NotImplemented

        return self.id == other.id

    def __ne__(self, other):
        """Check if the IDs of the Notifications are not equal."""
        return not self == other



from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from Weverse.models import Post


class Photo:
    r"""A Photo object that represents a Weverse Photo that belongs to media or a post.

    It is not suggested to create a Photo manually, but rather through the
    following method: :class:`Weverse.objects.create_photo_objects`

    The information retrieved on a Photo is directly from the Weverse API and altered to fit this class.

    .. container:: operations

        .. describe:: x == y

            Checks if two Photo objects have the same ID.

        .. describe:: x != y

            Checks if two Photo objects do not have the same ID.

        .. describe:: str(x)

            Returns the file name.

    Parameters
    ----------
    id: int
        The ID of the photo.
    content_index: int
        Index the photo is in from a bundle of photos.
    thumbnail_img_url: str
        The thumbnail image link.
    thumbnail_img_width: str
        The original image width.
    thumbnail_img_height: str
        The thumbnail image height.
    original_img_url: str
        The original image link.
    original_img_width: str
        The original image width.
    original_img_height: str
        The original image height.
    file_name: str
        File name of the photo.

    Attributes
    -----------
    id: int
        The ID of the photo.
    media_id: Optional[int]
        The media ID of the photo (if there is one).
    content_index: int
        Index the photo is in from a bundle of photos.
    thumbnail_img_url: str
        The thumbnail image link.
    thumbnail_img_width: str
        The original image width.
    thumbnail_img_height: str
        The thumbnail image height.
    original_img_url: str
        The original image link.
    original_img_width: str
        The original image width.
    original_img_height: str
        The original image height.
    file_name: str
        File name of the photo.
    post: Post
        The Post Object the photo belongs to.
    """
    def __init__(self, **kwargs):
        self.id = kwargs.get('photo_id')
        self.media_id = kwargs.get('media_id')
        self.content_index = kwargs.get('content_index')
        self.thumbnail_img_url = kwargs.get('thumbnail_img_url')
        self.thumbnail_img_width = kwargs.get('thumbnail_img_width')
        self.thumbnail_img_height = kwargs.get('thumbnail_img_height')
        self.original_img_url = kwargs.get('original_img_url')
        self.original_img_width = kwargs.get('original_img_width')
        self.original_img_height = kwargs.get('original_img_height')
        self.file_name = kwargs.get('file_name')
        self.post: Optional[Post] = None

    def __eq__(self, other):
        """Check if the IDs of the Photo objects are equal."""
        if not isinstance(other, Photo):
            return NotImplemented

        return self.id == other.id

    def __ne__(self, other):
        """Check if the IDs of the Photo objects are not equal."""
        return not self == other

    def __str__(self):
        """Returns the file name."""
        return f"{self.file_name}"



import asyncio
import concurrent.futures
import functools
import os
from typing import Optional, List, Union
from os import system as terminal
import aiohttp
import aiofiles
from asyncio import get_event_loop
from .models import Community, Post as w_Post, Notification, Announcement, Media, VideoStream
from . import WeverseClient, create_post_objects, create_community_objects, create_notification_objects, \
    create_comment_objects, create_media_object, iterate_community_media_categories, create_announcement_object, \
    InvalidCredentials, LoginFailed, InvalidToken, NoHookFound, check_expired_token, create_video_objects
from json import dumps as dumps_


VERBOSE = False

class WeverseClientAsync(WeverseClient):
    r"""
    Asynchronous Weverse Client that Inherits from :ref:`WeverseClient`.

    Parameters
    ----------
    loop:
        Asyncio Event Loop
    kwargs:
        Same as :ref:`WeverseClient`.


    Attributes
    -----------
    loop:
        Asyncio Event Loop

    Attributes are the same as :ref:`WeverseClient`.
    """

    def __init__(self, loop=get_event_loop(), **kwargs):
        self._follow_new_communities = True
        self._time_passed: int = 0
        self._request_payload_for_follow = {
            "profileNickname": "NONE",
            "profileImgPath": "https://cdn-contents.weverse.io/static/profile/profile_defalut_img_05.png"
        }
        self.loop = loop
        self._cookies: Optional[dict] = None
        self.__cookies_test_url = "https://weversewebapi.weverse.io/wapi/v1/communities/2/videos/4093"
        super().__init__(**kwargs)

        if self.verbose:
            global VERBOSE
            VERBOSE = self.verbose

    async def start(self, create_old_posts=False, create_notifications=True, create_media=False,
                    follow_new_communities=True):
        """Creates internal cache.

        This is the main process that should be run.

        This is a coroutine and must be awaited.

        :parameter create_old_posts: (:class:`bool`) Whether to create cache for old posts.
        :parameter create_notifications: (:class:`bool`) Whether to create/update cache for old notifications.
        :parameter create_media: (:class:`bool`) Whether to create/update cache for old media.
        :param follow_new_communities: bool
            Check for new communities and automatically follow them.

        :raises: :class:`Weverse.error.InvalidToken`
            If the token was invalid.
        :raises: :class:`Weverse.error.BeingRateLimited`
            If the client is being rate-limited.
        :raises: :class:`Weverse.error.LoginFailed`
            Login process had failed.
        :raises: :class:`asyncio.exceptions.TimeoutError`
            Waited too long for a login.
        :raises: :class:`Weverse.error.InvalidCredentials`
            If the user credentials were invalid or not provided.
        """
        try:
            if not self.web_session:
                self.web_session = aiohttp.ClientSession()

            if not self._login_info_exists and not self._token_exists:
                raise InvalidCredentials

            if self._login_info_exists:
                await self._try_login()
                await self._wait_for_login()  # wait for login or an exception to occur.

            if not await self.check_token_works():
                raise InvalidToken

            # create all communities that are subscribed to
            await self.create_communities()  # communities should be created no matter what

            # create and update community artists and their tabs
            await self.create_community_artists_and_tabs()

            self._follow_new_communities = follow_new_communities
            if self._follow_new_communities:
                await self.follow_all_communities()

            # create and update user notifications
            if create_notifications:
                await self.get_user_notifications()

            for community in self.all_communities.values():
                # load up posts
                if create_old_posts:
                    await self.create_posts(community)

                # load up media
                if create_media:
                    await self.create_media(community)

            self.cache_loaded = True

            if self._hook:
                if self.verbose:
                    print("Starting Notification Loop for Weverse Client.")
                await self._start_loop_for_hook()
        except Exception as err:
            raise err

    async def _start_loop_for_hook(self):
        """
        Start checking for new notifications in a new loop and call the hook with the list of new Notifications
        This will also create the posts associated with the notification so they can be used efficiently.
        This is a coroutine and must be awaited.
        """
        if not self._hook:
            raise NoHookFound

        self._hook_loop = True
        while self._hook_loop:
            if self._follow_new_communities:
                if self._time_passed == 14400:  # 4 hours in seconds
                    self._time_passed = 0
                    await self.follow_all_communities()

                self._time_passed += 30
            await asyncio.sleep(30)
            new_notifications = await self.update_cache_from_notification()
            if not new_notifications:
                continue

            if not asyncio.iscoroutinefunction(self._hook):
                self._hook(new_notifications)
            else:
                await self._hook(new_notifications)

    async def _try_login(self):
        """
        Will attempt to login to Weverse and set refresh token and token.
        This is a coroutine and must be awaited.
        """
        self._login(self.__process_login)

    async def __process_login(self, login_payload: dict):
        """
        Will process login credentials and set refresh token and token.
        This is a coroutine and must be awaited.
        Parameters
        ----------
        login_payload: dict
            The client's login payload
        """
        async with self.web_session.post(url=self._login_url, json=login_payload) as resp:
            if self.check_status(resp.status, self._login_url):
                data = await resp.json()
                refresh_token = data.get("refresh_token")
                token = data.get("access_token")
                if refresh_token:
                    self._set_refresh_token(refresh_token)
                if token:
                    self._set_token(token)
                self.expired_token = False
                return
        self._set_exception(LoginFailed())

    async def _refresh_token(self):
        """
        Refresh a token while logged in.

        This is a coroutine and must be awaited.

        """
        async with self.web_session.post(url=self._login_url, json=self._refresh_payload) as resp:
            if self.check_status(resp.status, self._login_url):
                data = await resp.json()
                token = data.get("access_token")
                if token:
                    self._set_token(token)
                self.expired_token = False
                return
        await self._try_login()
        await self._wait_for_login()

    @check_expired_token
    async def create_media(self, community: Community):
        """Paginate through a community's media and add it to object cache.

        :parameter community: :ref:`Community` the posts exist under.
        """
        media_tab_url = f"{self._api_stream_url}{community.id}/{self._api_media_tab}"
        async with self.web_session.get(media_tab_url, headers=self._headers) as resp:
            if self.check_status(resp.status, media_tab_url):
                data = await resp.json()
                media_objects, photo_media_dicts = iterate_community_media_categories(data)

                # This endpoint does NOT give us any information about the photos, therefore we must make
                # a separate api call to retrieve proper photo information for the photo media.
                for media in photo_media_dicts:
                    media_obj = await self.fetch_media(community.id, media.get("id"))
                    if media_obj:
                        media_objects.append(media_obj)

                self._add_media_to_cache(media_objects)

    @check_expired_token
    async def create_communities(self):
        """Get and Create the communities the logged in user has access to.

        This is a coroutine and must be awaited.
        """
        async with self.web_session.get(self._api_communities_url, headers=self._headers) as resp:
            if self.check_status(resp.status, self._api_communities_url):
                data = await resp.json()
                user_communities = data.get("communities")
                self.all_communities = create_community_objects(user_communities, self.all_communities)

    @check_expired_token
    async def create_community_artists_and_tabs(self, specific_community_ids: List[int] = None):
        """Create the community artists and tabs and add them to their respective communities.

        :param specific_community_ids: List[int]
            Will only do this list of community ids from the already existing communities.

        This is a coroutine and must be awaited.
        """
        for community in self.all_communities.values():
            if specific_community_ids:
                if community.id not in specific_community_ids:
                    continue
            url = self._api_communities_url + str(community.id)
            async with self.web_session.get(url, headers=self._headers) as resp:
                if self.check_status(resp.status, url):
                    data = await resp.json()
                    self.process_community_artists_and_tabs(community, data)
                    for artist in community.artists:
                        self.all_artists[artist.id] = artist
                    for tab in community.tabs:
                        self.all_tabs[tab.id] = tab

    @check_expired_token
    async def create_posts(self, community: Community, next_page_id: int = None):
        """Paginate through a community's posts and add it to object cache.

        This is a coroutine and must be awaited.

        :parameter community: :ref:`Community` the posts exist under.
        :parameter [OPTIONAL] next_page_id: Next Page ID (Weverse paginates posts).
        """
        artist_tab_url = self._api_communities_url + str(community.id) + '/' + self._api_all_artist_posts_url
        if next_page_id:
            artist_tab_url = artist_tab_url + "?from=" + str(next_page_id)
        async with self.web_session.get(artist_tab_url, headers=self._headers) as resp:
            if self.check_status(resp.status, artist_tab_url):
                data = await resp.json()
                posts = create_post_objects(data.get('posts'), community)
                for post in posts:
                    self.all_posts[post.id] = post
                    if post.photos:
                        for photo in post.photos:
                            self.all_photos[photo.id] = photo

                    if post.videos:
                        for video in post.photos:
                            self.all_videos[video.video_url] = video

                if not data.get('isEnded'):
                    await self.create_posts(community, data.get('lastId'))

    @check_expired_token
    async def create_post(self, community: Community, post_id) -> w_Post:
        """Create a post and update the cache with it. This is meant for an individual post.

        This is a coroutine and must be awaited.

        :parameter community: :ref:`Community` the post was created under.
        :parameter post_id: The id of the post we are needing to fetch.
        """
        post_url = self._api_communities_url + str(community.id) + '/posts/' + str(post_id)
        async with self.web_session.get(post_url, headers=self._headers) as resp:
            if self.check_status(resp.status, post_url):
                data = await resp.json()
                return (create_post_objects([data], community, new=True))[0]

    @check_expired_token
    async def get_user_notifications(self):
        """Get a list of updated user notification objects.

        This is a coroutine and must be awaited.

        :returns: List[:ref:`Notification`]
        """
        self._old_notifications = self.user_notifications  # important for keeping track of what is new.

        async with self.web_session.get(self._api_notifications_url, headers=self._headers) as resp:
            if self.check_status(resp.status, self._api_notifications_url):
                data = await resp.json()
                self.user_notifications = create_notification_objects(data.get('notifications'))
                for user_notification in self.user_notifications:
                    self.all_notifications[user_notification.id] = user_notification
                return self.user_notifications

    @check_expired_token
    async def check_new_user_notifications(self) -> bool:
        """Checks if there is a new user notification, updates the cache, and returns if there was.

        This is a coroutine and must be awaited.

        :returns: (:class:`bool`) Whether there is a new notification.


        This endpoint has been acting a bit off and not producing accurate results. It would be recommended to
        instantly get new notifications with :ref:`update_cache_from_notification` instead.
        """
        async with self.web_session.get(self._api_new_notifications_url, headers=self._headers) as resp:
            if self.check_status(resp.status, self._api_new_notifications_url):
                data = await resp.json()
                has_new = data.get('has_new')
                if has_new:
                    # update cache
                    # Not that cache_loaded necessarily matters here,
                    # but just in case other checks are happening concurrently.
                    self.cache_loaded = False
                    await self.update_cache_from_notification()
                    self.cache_loaded = True
                return has_new

    @check_expired_token
    async def translate(self, post_or_comment_id, is_post=False, is_comment=False, p_obj=None, community_id=None):
        """Translates a post or comment, must set post or comment to True.

        This is a coroutine and must be awaited.

        :parameter post_or_comment_id: A post or comment ID.
        :parameter [OPTIONAL] is_post: If we passed in a post.
        :parameter [OPTIONAL] is_comment: If we passed in a comment
        :parameter [OPTIONAL] p_obj: The object we are looking to translate
        :parameter [OPTIONAL] community_id: The community id the post/comment was made under.
        :returns: (:class:`str`) Translated message or NoneType
        """
        post_check = False
        comment_check = False
        method_url = None
        if is_post:
            method_url = "posts/"
            if not p_obj:
                p_obj = self.get_post_by_id(post_or_comment_id)
            post_check = True
        elif is_comment:
            method_url = "comments/"
            if not p_obj:
                p_obj = self.get_comment_by_id(post_or_comment_id)
            comment_check = True
        if not community_id:
            if p_obj:
                if comment_check:
                    if p_obj.post:
                        community_id = p_obj.post.artist.community_id
                if post_check:
                    if p_obj.artist:
                        community_id = p_obj.artist.community_id
            else:
                return None
        url = self._api_communities_url + str(community_id) + "/" + method_url + str(
            post_or_comment_id) + "/translate?languageCode=en"
        async with self.web_session.get(url, headers=self._headers) as resp:
            if self.check_status(resp.status, url):
                data = await resp.json()
                return data.get('translation')

    @check_expired_token
    async def fetch_artist_comments(self, community_id, post_id):
        """Fetches the artist comments on a post.

        This is a coroutine and must be awaited.

        :parameter community_id: Community ID the post is on.
        :parameter post_id: Post ID to fetch the artist comments of.
        :returns: List[:ref:`Comment`]
        """
        post_comments_url = self._api_communities_url + str(community_id) + '/posts/' + str(post_id) + "/comments/"
        async with self.web_session.get(post_comments_url, headers=self._headers) as resp:
            if self.check_status(resp.status, post_comments_url):
                data = await resp.json()
                return create_comment_objects(data.get('artistComments'))

    @check_expired_token
    async def fetch_comment_body(self, community_id, comment_id) -> str:
        """Fetches a comment from its ID.

        This is a coroutine and must be awaited.

        :parameter community_id: The ID of the community the comment belongs to.
        :parameter comment_id: The ID of the comment to fetch.
        :returns: (:class:`str`) Body of the comment.
        """
        comment_url = f"{self._api_communities_url}{str(community_id)}/comments/{comment_id}/"
        async with self.web_session.get(comment_url, headers=self._headers) as resp:
            if self.check_status(resp.status, comment_url):
                data = await resp.json()
                return data.get('body')

    @check_expired_token
    async def fetch_media(self, community_id, media_id) -> Optional[Media]:
        """Receive media object based on media id.

        This is a coroutine and must be awaited.

        :parameter community_id: The ID of the community the media belongs to.
        :parameter media_id: The ID of the media to fetch.
        :returns: :ref:`Media` or NoneType
        """
        media_url = self._api_communities_url + str(community_id) + "/medias/" + str(media_id)
        async with self.web_session.get(media_url, headers=self._headers) as resp:
            if self.check_status(resp.status, media_url):
                data = await resp.json()
                return create_media_object(data.get('media'))

    @check_expired_token
    async def fetch_announcement(self, community_id: int, announcement_id: int) -> Optional[Announcement]:
        """Receive announcement object based on announcement id.

        This is a coroutine and must be awaited.

        :parameter community_id: The ID of the community the media belongs to.
        :parameter announcement_id: The ID of the announcement to fetch.
        :returns: :ref:`Announcement` or NoneType
        """
        announcement_url = self._api_communities_url + str(community_id) + "/notices/" + str(announcement_id)
        async with self.web_session.get(announcement_url, headers=self._headers) as resp:
            if self.check_status(resp.status, announcement_url):
                data = await resp.json()
                return create_announcement_object(data)

    @staticmethod
    def __generate_random_nickname():
        from random import randint
        nickname_length = randint(10, 20)
        nickname = ""
        for i in range(0, nickname_length + 1):
            nickname += chr(randint(0, 255))
        return nickname

    @check_expired_token
    async def follow_community(self, community_id: Union[int, str], attempts: int = 0):
        r"""
        Follow a community

        :param community_id: Union[int, str]
            The community ID to follow.
        :param attempts: int
            The number of attempts for choosing a nickname after error.

        """
        url = self._api_communities_url + str(community_id)
        self._request_payload_for_follow['profileNickname'] = self.__generate_random_nickname()
        _headers = self._headers
        _headers['Content-Type'] = 'application/json'
        async with self.web_session.put(url, headers=_headers, data=dumps_(self._request_payload_for_follow)) as \
                resp:
            if resp.status == 400 and attempts < 1:
                return await self.follow_community(community_id, attempts + 1)
            if self.check_status(resp.status, url):
                if self.verbose:
                    print(f"Followed {community_id}.")

    @check_expired_token
    async def follow_all_communities(self):
        r"""Follow all communities on Weverse"""
        communities_to_follow = [community_id for community_id in await self.get_all_community_ids() if
                                 community_id not in
                                 self.all_communities]
        for community_id in communities_to_follow:
            await self.follow_community(community_id)

        await self.create_communities()
        await self.create_community_artists_and_tabs(specific_community_ids=communities_to_follow)

    @check_expired_token
    async def get_all_community_ids(self) -> List[int]:
        r"""
        Get all the communities on Weverse.

        :returns: List[int]
            A list of community ids
        """
        url = self._api_url + 'app-properties/key/webCommunityRedirectPath'
        async with self.web_session.get(url, headers=self._headers) as resp:
            if self.check_status(resp.status, url):
                list_of_communities_: dict = await resp.json()
                return [community_info['id'] for community_info in list_of_communities_.get('communities')]
        return []

    async def update_cache_from_notification(self) -> List[Notification]:
        """Grab a new post based from new notifications and add it to cache.

        Will also return the new notifications found.

        This is a coroutine and must be awaited.

        :returns: List[:class:`models.Notification`]
        """
        new_notifications = []
        try:
            notifications = await self.get_user_notifications()

            if not notifications:
                return new_notifications

            new_notifications = self.get_new_notifications()
            for notification in new_notifications:
                await self.__manage_notification_posts(notification)
        except Exception as e:
            if self.verbose:
                print(f"Failed to update Weverse Cache - {e}")
        return new_notifications

    async def __manage_notification_posts(self, notification: Notification):
        """Manages the creation of Notification posts and comments.

        This is a coroutine and must be awaited.

        :param notification: Notification to create comments and posts for.
        """
        notification_type = self.determine_notification_type(notification.message)
        community = self.get_community_by_id(notification.community_id)
        if notification_type == 'comment':
            artist_comments = await self.fetch_artist_comments(notification.community_id, notification.contents_id)
            if artist_comments:
                comment = artist_comments[0]
                comment.post = self.get_post_by_id(comment.post_id)
                if comment.post:
                    if comment.post.artist_comments:
                        comment.post.artist_comments.insert(0, comment)
                    else:
                        comment.post.artist_comments = [comment]
                self.all_comments[comment.id] = comment
        elif notification_type in ["tofans", "post"]:
            post = await self.create_post(community, notification.contents_id)
            if post:
                self.all_posts[post.id] = post
        elif notification_type == 'media':
            media = await self.fetch_media(community.id, notification.contents_id)
            if media:
                self.all_media[media.id] = media
        elif notification_type == 'announcement':
            announcement = await self.fetch_announcement(community.id, notification.contents_id)
            if announcement:
                self.all_announcements[announcement.id] = announcement

    async def check_token_works(self) -> bool:
        """
        Check if a token is invalid.

        This is a coroutine and must be awaited.

        :returns: (:class:`bool`) True if the token works.
        """
        async with self.web_session.get(url=self._user_endpoint, headers=self._headers) as resp:
            self._expired_token = not resp.status == 200
            return not self._expired_token

    @property
    async def cookies(self) -> Optional[dict]:
        """Get the user's cookies in order to access media."""
        if self._cookies:
            return self._cookies
        else:
            return await self.get_cookies(video_url_without_drm_type=self.__cookies_test_url)

    async def get_cookies(self, video_url_without_drm_type) -> Optional[dict]:
        """Get the user's cookies in order to access media.

        :param video_url_without_drm_type:
            EX: https://weversewebapi.weverse.io/wapi/v1/communities/2/videos/4093

        :returns: Optional[dict]
            A dictionary containing a signed cookie.
        """

        if not self._cookies:
            url = video_url_without_drm_type + '?drmType=Widevine'
            async with self.web_session.get(url=url, headers=self._headers) as resp:
                if self.check_status(resp.status, url):
                    data = await resp.json()
                    self._cookies = data['signedCookie']
        return self._cookies

    async def download_video_stream(self, video_stream_obj: VideoStream, output_file_path):
        """
        Download a video stream to a local folder.

        Parameters
        ----------
        video_stream_obj: :ref:`VideoStream`
        output_file_path: str
            Full file path with file extension.

        Returns
        -------
        Returns False if no community id is found with the :ref:`VideoStream` object.
        """
        if not video_stream_obj.community_id:
            return False

        url = f"{self._api_communities_url}{video_stream_obj.community_id}/videos/{video_stream_obj.video_id}"
        cookies = await self.get_cookies(url)
        self._headers['cookie'] = cookies
        for m3u8_url in video_stream_obj.m3u8_urls:
            async with self.web_session.get(url=m3u8_url,
                                            headers=self._headers) as resp:
                if not self.check_status(resp.status, m3u8_url):
                    continue

                data_bytes_string = await resp.read()
                data_string = data_bytes_string.decode('ascii')
                lines = data_string.split("\n")
                ts_file_names = [line for line in lines if line.endswith('.ts')]
                ts_file_urls = [f"{video_stream_obj.base_url}{ts_file_name}" for ts_file_name in ts_file_names]
                ts_file_paths = [f"./{ts_file_name}" for ts_file_name in ts_file_names]
                downloaded_list = await self._download_ts_files(ts_file_urls, ts_file_paths)
                concat_files_syntax = '|'.join(downloaded_list)
                ffmpeg_concat_protocol = f'ffmpeg -i "concat:{concat_files_syntax}" -c copy {output_file_path}'
                await self.run_blocking_code(self._run_in_terminal, ffmpeg_concat_protocol)
                await self.run_blocking_code(self._remove_files, downloaded_list)
                break  # we have our output file for highest quality found.

    async def _download_ts_files(self, urls, file_paths) -> List[str]:
        """
        Download TS files.

        Parameters
        ----------
        urls: List[str]
            A list of links. MUST equal the length of file paths.
        file_paths: List[str]
            File paths to download. MUST equal the length of urls.

        :returns: List[str]
            Returns a list of downloaded file paths.

        """
        downloaded_files = []
        for idx, url in enumerate(urls):
            async with self.web_session.get(url, headers=self._headers) as resp:
                if self.check_status(resp.status, url):
                    async with aiofiles.open(file_paths[idx], mode='wb') as file:
                        await file.write(await resp.read())
                        downloaded_files.append(file_paths[idx])
        return downloaded_files

    async def run_blocking_code(self, funcs, *args, **kwargs) -> list:
        """Run blocking code safely in a new thread.
        DO NOT pass in an asynchronous function. If an asynchronous function has blocking code, the event loop will
        also block. There were several attempts made to make it compatible with asynchronous functions, but it was a
        headache to work with.

        :param funcs: The blocking function that needs to be called.
            May also pass in a list of functions
            with the 0th index as the callable function,
            the 1st index as the args for that function,
            and the 2nd index as the kwargs for that function.
        :param args: The args to pass into the blocking function.
        :param kwargs: The keyword args to pass into the blocking function.
        :returns: List of results in no particular order. Make sure the output can be managed with no specific order.
        """
        loop = asyncio.get_running_loop()
        try:
            with concurrent.futures.ThreadPoolExecutor(max_workers=5) as pool:
                results = []  # a list of results
                if not isinstance(funcs, list):
                    funcs = [[funcs, args, kwargs]]

                for func in funcs:
                    callable_function = func[0]
                    func_args = func[1]
                    func_kwargs = func[2]
                    if callable(callable_function):
                        results.append(await loop.run_in_executor(pool, functools.partial(callable_function,
                                                                                          *func_args, **func_kwargs)))
                return results
        except Exception as e:
            if self.verbose:
                print(f"Exception - {e} - {func} - {args} - {kwargs}")
        return []

    @staticmethod
    def _run_in_terminal(command):
        """Run a command in the terminal. Will Block unless executed from :ref:`run_blocking_code`"""
        terminal(command)

    @staticmethod
    def _remove_files(file_paths: List[str]):
        """
        Removes a list of files. Will Block unless executed from :ref:`run_blocking_code`

        Parameters
        ----------
        file_paths: List[str]
            A list of full file paths
        """
        for file_path in file_paths:
            try:
                os.unlink(file_path)
            except Exception as e:
                if VERBOSE:
                    print(f"{file_path} could not be unlinked. - {e}")



from typing import Optional
import re


class Announcement:
    r"""An Announcement object that represents a Weverse Notice for a Community.

    It is not suggested to create an Announcement manually, but rather through the
    following method: :class:`Weverse.objects.create_announcement_objects`

    The information retrieved on a Post is directly from the Weverse API and altered to fit this class.

    .. container:: operations

        .. describe:: x == y

            Checks if two Announcement objects have the same ID.

        .. describe:: x != y

            Checks if two Announcement objects do not have the same ID.

        .. describe:: str(x)

            Returns the Announcement content.

    Parameters
    ----------
    id: int
        The ID of the post.
    communityId: int
        The Community ID.
    title: str
        The title of the announcement notice.
    content: str
        The HTML body of the page notice.
    createdAt: str
        Timestamp with the date of when the announcement was created.
    exposedAt: str
        Timestamp with the date of when the announcement was released.
    categoryId: int
        Category that the announcement belongs to (used for paginating or quick endpoint access)
    fcOnly: bool
        If only premium members have access to the announcement.

    Attributes
    -----------
    id: int
        The ID of the post.
    community_id: int
        The Community ID.
    title: str
        The title of the announcement notice.
    html_content: str
        The HTML body of the page notice.
    created_at: str
        Timestamp with the date of when the announcement was created.
    exposed_at: str
        Timestamp with the date of when the announcement was released.
    category_id: int
        Category that the announcement belongs to (used for paginating or quick endpoint access)
    fc_only: bool
        If only premium members have access to the announcement.
    image_url: Optional[str]
        An image url if one is present.
    content: str
        Body Content without the HTML tags.

    """
    def __init__(self, **kwargs):
        self.id = kwargs.get("id")
        self.community_id = kwargs.get("communityId")
        self.title = kwargs.get("title")
        self.html_content: str = kwargs.get("content")
        self.created_at = kwargs.get("createdAt")
        self.exposed_at = kwargs.get("exposedAt")
        self.category_id = kwargs.get("categoryId")
        self.fc_only: bool = kwargs.get("fcOnly")
        self.image_url: Optional[str] = self.__get_image()
        self.content = self.__remove_html()

    def __eq__(self, other):
        """Check if the IDs of the Announcement objects are equal."""
        if not isinstance(other, Announcement):
            return NotImplemented

        return self.id == other.id

    def __ne__(self, other):
        """Check if the IDs of the Announcement objects are not equal."""
        return not self == other

    def __str__(self):
        """Returns the Announcement content without the HTML tags."""
        return f"{self.content}"

    def __get_image(self):
        """Retrieve the image url from the html content if any exists."""
        src_loc = self.html_content.find("src=")
        if src_loc == -1:
            return ""
        skipped_start = self.html_content[src_loc:len(self.html_content)]
        end_of_src = skipped_start.find(" ")  # a space will indicate the end of the src
        if skipped_start == -1:
            return ""

        image_src = skipped_start[0:end_of_src]
        image_src = image_src.replace("src=", "")
        image_src = image_src.replace('\\', "")
        image_src = image_src.replace('"', "")
        return image_src

    def __remove_html(self):
        """Removes HTML tags of the html content and returns a finalized version."""
        content = self.html_content.replace("<br>", "\n")  # replace new line tags before they get replaced.
        html_cleaner = re.compile('<.*?>|&([a-z0-9]+|#[0-9]{1,6}|#x[0-9a-f]{1,6});')
        clean_text = re.sub(html_cleaner, '', content)
        return clean_text





import json
import time
from typing import Optional, List

import requests
from .models import Community, Post as w_Post, Notification, Announcement
from . import WeverseClient, create_post_objects, create_community_objects, create_notification_objects, \
    create_comment_objects, create_media_object, iterate_community_media_categories, create_announcement_object, \
    InvalidCredentials, LoginFailed, InvalidToken, NoHookFound, check_expired_token


class WeverseClientSync(WeverseClient):
    r"""
    Synchronous Weverse Client that Inherits from :ref:`WeverseClient`.

    Parameters
    ----------
    kwargs:
        Same as :ref:`WeverseClient`.


    Attributes are the same as :ref:`WeverseClient`.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def start(self, create_old_posts=False, create_notifications=True, create_media=False):
        """Creates internal cache.

        This is the main process that should be run.

        :parameter create_old_posts: (:class:`bool`) Whether to create cache for old posts.
        :parameter create_notifications: (:class:`bool`) Whether to create/update cache for old notifications.
        :parameter create_media: (:class:`bool`) Whether to create/update cache for old media.

        :raises: :class:`Weverse.error.InvalidToken`
            If the token was invalid.
        :raises: :class:`Weverse.error.BeingRateLimited`
            If the client is being rate-limited.
        :raises: :class:`Weverse.error.LoginFailed`
            Login process had failed.
        :raises: :class:`Weverse.error.InvalidCredentials`
            If the user credentials were invalid or not provided.
        """
        try:
            if not self.web_session:
                self.web_session = requests.Session()

            if not self._login_info_exists and not self._token_exists:
                raise InvalidCredentials

            if self._login_info_exists:
                self._try_login()

            if not self.check_token_works():
                raise InvalidToken

            # create all communities that are subscribed to
            self.create_communities()  # communities should be created no matter what

            # create and update community artists and their tabs
            self.create_community_artists_and_tabs()

            # create and update user notifications
            if create_notifications:
                self.get_user_notifications()

            for community in self.all_communities.values():
                # load up posts
                if create_old_posts:
                    self.create_posts(community)

                if create_media:
                    self.create_media(community)

            self.cache_loaded = True

            if self._hook:
                if self.verbose:
                    print("Starting Notification Loop for Weverse Client.")
                self._start_loop_for_hook()
        except Exception as err:
            raise err

    def _start_loop_for_hook(self):
        """
        Start checking for new notifications in a new loop and call the hook with the list of new Notifications
        This will also create the posts associated with the notification so they can be used efficiently.
        """
        if not self._hook:
            raise NoHookFound

        self._hook_loop = True
        while self._hook_loop:
            time.sleep(30)
            new_notifications = self.update_cache_from_notification()
            if not new_notifications:
                continue

            self._hook(new_notifications)

    def _try_login(self):
        """
        Will attempt to login to Weverse and set refresh token and token.
        """
        self._login(self.__process_login)

    def __process_login(self, login_payload: dict):
        """
        Will process login credentials and set refresh token and token.
        Parameters
        ----------
        login_payload: dict
            The client's login payload
        """
        with self.web_session.post(url=self._login_url, json=login_payload) as resp:
            if self.check_status(resp.status_code, self._login_url):
                data = json.loads(resp.text)
                refresh_token = data.get("refresh_token")
                token = data.get("access_token")
                if refresh_token:
                    self._set_refresh_token(refresh_token)
                if token:
                    self._set_token(token)
                self.expired_token = False
                return
        raise LoginFailed

    def _refresh_token(self):
        """
        Refresh a token while logged in.
        """
        with self.web_session.post(url=self._login_url, json=self._refresh_payload) as resp:
            if self.check_status(resp.status_code, self._login_url):
                data = json.loads(resp.text)
                token = data.get("access_token")
                if token:
                    self._set_token(token)
                self.expired_token = False
                return
        self._set_refresh_token("")  # resetting refresh token.
        self._try_login()  # will attempt to log in again.

    @check_expired_token
    def _check_new_notifications(self) -> List[Notification]:
        """
        Checks and returns new notifications.
        Compares with the already existing notifications.
        This will also create the posts associated with the notification so they can be used efficiently.
        This is a coroutine and must be awaited.

        Returns
        -------
        A list of new Notifications.: List[:class:`models.Notification`]
        """
        all_new_notifications = []

        if self.check_new_user_notifications():

            all_new_notifications = self.get_new_notifications()
        return all_new_notifications

    @check_expired_token
    def create_media(self, community: Community):
        """Paginate through a community's media and add it to object cache.

        :parameter community: :ref:`Community` the posts exist under.
        """
        media_tab_url = f"{self._api_stream_url}{community.id}/{self._api_media_tab}"
        with self.web_session.get(media_tab_url, headers=self._headers) as resp:
            if self.check_status(resp.status_code, media_tab_url):
                response_text_as_dict = json.loads(resp.text)
                media_objects, photo_media_dicts = iterate_community_media_categories(response_text_as_dict)

                # This endpoint does NOT give us any information about the photos, therefore we must make
                # a separate api call to retrieve proper photo information for the photo media.
                for media in photo_media_dicts:
                    media_obj = self.fetch_media(community.id, media.get("id"))
                    if media_obj:
                        media_objects.append(media_obj)

                self._add_media_to_cache(media_objects)

    @check_expired_token
    def create_communities(self):
        """Get and Create the communities the logged in user has access to."""
        with self.web_session.get(self._api_communities_url, headers=self._headers) as resp:
            if self.check_status(resp.status_code, self._api_communities_url):
                response_text = resp.text
                response_text_as_dict = json.loads(response_text)
                user_communities = response_text_as_dict.get("communities")
                self.all_communities = create_community_objects(user_communities)

    @check_expired_token
    def create_community_artists_and_tabs(self):
        """Create the community artists and tabs and add them to their respective communities."""
        for community in self.all_communities.values():
            url = self._api_communities_url + str(community.id)
            with self.web_session.get(url, headers=self._headers) as resp:
                if self.check_status(resp.status_code, url):
                    response_text = resp.text
                    response_text_as_dict = json.loads(response_text)
                    self.process_community_artists_and_tabs(community, response_text_as_dict)
                    for artist in community.artists:
                        self.all_artists[artist.id] = artist
                    for tab in community.tabs:
                        self.all_tabs[tab.id] = tab

    @check_expired_token
    def create_posts(self, community: Community, next_page_id: int = None):
        """Paginate through a community's posts and add it to object cache.

        :parameter community: :ref:`Community` the posts exist under.
        :parameter [OPTIONAL] next_page_id: Next Page ID (Weverse paginates posts).
        """
        artist_tab_url = self._api_communities_url + str(community.id) + '/' + self._api_all_artist_posts_url
        if next_page_id:
            artist_tab_url = artist_tab_url + "?from=" + str(next_page_id)
        with self.web_session.get(artist_tab_url, headers=self._headers) as resp:
            if self.check_status(resp.status_code, artist_tab_url):
                response_text = resp.text
                response_text_as_dict = json.loads(response_text)
                posts = create_post_objects(response_text_as_dict.get('posts'), community)
                for post in posts:
                    self.all_posts[post.id] = post
                    if post.photos:
                        for photo in post.photos:
                            self.all_photos[photo.id] = photo

                    if post.videos:
                        for video in post.photos:
                            self.all_videos[video.video_url] = video
                if not response_text_as_dict.get('isEnded'):
                    self.create_posts(community, response_text_as_dict.get('lastId'))

    @check_expired_token
    def create_post(self, community: Community, post_id) -> w_Post:
        """Create a post and update the cache with it. This is meant for an individual post.

        :parameter community: :ref:`Community` the post was created under.
        :parameter post_id: The id of the post we are needing to fetch.
        """
        post_url = self._api_communities_url + str(community.id) + '/posts/' + str(post_id)
        with self.web_session.get(post_url, headers=self._headers) as resp:
            if self.check_status(resp.status_code, post_url):
                response_text = resp.text
                response_text_as_dict = json.loads(response_text)
                return (create_post_objects([response_text_as_dict], community, new=True))[0]

    @check_expired_token
    def get_user_notifications(self):
        """Get a list of updated user notification objects.

        :returns: List[:ref:`Notification`]
        """
        with self.web_session.get(self._api_notifications_url, headers=self._headers) as resp:
            if self.check_status(resp.status_code, self._api_notifications_url):
                response_text = resp.text
                response_text_as_dict = json.loads(response_text)
                self.user_notifications = create_notification_objects(response_text_as_dict.get('notifications'))
                for user_notification in self.user_notifications:
                    self.all_notifications[user_notification.id] = user_notification
                return self.user_notifications

    @check_expired_token
    def check_new_user_notifications(self):
        """Checks if there is a new user notification, updates the cache, and returns if there was.

        :returns: (:class:`bool`) Whether there is a new notification.

        This endpoint has been acting a bit off and not producing accurate results. It would be recommended to
        instantly get new notifications with :ref:`update_cache_from_notification` instead.
        """
        with self.web_session.get(self._api_new_notifications_url, headers=self._headers) as resp:
            if self.check_status(resp.status_code, self._api_new_notifications_url):
                response_text = resp.text
                response_text_as_dict = json.loads(response_text)
                has_new = response_text_as_dict.get('has_new')
                if has_new:
                    # update cache
                    # Not that cache_loaded necessarily matters here,
                    # but just in case other checks are happening concurrently.
                    self.cache_loaded = False
                    self.update_cache_from_notification()
                    self.cache_loaded = True
                return has_new

    @check_expired_token
    def translate(self, post_or_comment_id, is_post=False, is_comment=False, p_obj=None, community_id=None):
        """Translates a post or comment, must set post or comment to True.

        :parameter post_or_comment_id: A post or comment ID.
        :parameter [OPTIONAL] is_post: If we passed in a post.
        :parameter [OPTIONAL] is_comment: If we passed in a comment
        :parameter [OPTIONAL] p_obj: The object we are looking to translate
        :parameter [OPTIONAL] community_id: The community id the post/comment was made under.
        :returns: (:class:`str`) Translated message or NoneType
        """
        post_check = False
        comment_check = False
        method_url = None
        if is_post:
            method_url = "posts/"
            if not p_obj:
                p_obj = self.get_post_by_id(post_or_comment_id)
            post_check = True
        elif is_comment:
            method_url = "comments/"
            if not p_obj:
                p_obj = self.get_comment_by_id(post_or_comment_id)
            comment_check = True
        if not community_id:
            if p_obj:
                if comment_check:
                    if p_obj.post:
                        community_id = p_obj.post.artist.community_id
                if post_check:
                    if p_obj.artist:
                        community_id = p_obj.artist.community_id
            else:
                return None
        url = self._api_communities_url + str(community_id) + "/" + method_url + str(
            post_or_comment_id) + "/translate?languageCode=en"
        with self.web_session.get(url, headers=self._headers) as resp:
            if self.check_status(resp.status_code, url):
                response_text = resp.text
                response_text_as_dict = json.loads(response_text)
                return response_text_as_dict.get('translation')

    @check_expired_token
    def fetch_artist_comments(self, community_id, post_id):
        """Fetches the artist comments on a post.

        :parameter community_id: Community ID the post is on.
        :parameter post_id: Post ID to fetch the artist comments of.
        :returns: List[:ref:`Comment`]
        """
        post_comments_url = self._api_communities_url + str(community_id) + '/posts/' + str(post_id) + "/comments/"
        with self.web_session.get(post_comments_url, headers=self._headers) as resp:
            if self.check_status(resp.status_code, post_comments_url):
                response_text = resp.text
                response_text_as_dict = json.loads(response_text)
                return create_comment_objects(response_text_as_dict.get('artistComments'))

    @check_expired_token
    def fetch_comment_body(self, community_id, comment_id):
        """Fetches a comment from its ID.

        :parameter community_id: The ID of the community the comment belongs to.
        :parameter comment_id: The ID of the comment to fetch.
        :returns: (:class:`str`) Body of the comment.
        """
        comment_url = f"{self._api_communities_url}{str(community_id)}/comments/{comment_id}/"
        with self.web_session.get(comment_url, headers=self._headers) as resp:
            if self.check_status(resp.status_code, comment_url):
                response_text = resp.text
                response_text_as_dict = json.loads(response_text)
                return response_text_as_dict.get('body')

    @check_expired_token
    def fetch_media(self, community_id, media_id):
        """Receive media object based on media id.

        :parameter community_id: The ID of the community the media belongs to.
        :parameter media_id: The ID of the media to fetch.
        :returns: :ref:`Media` or NoneType
        """
        media_url = self._api_communities_url + str(community_id) + "/medias/" + str(media_id)
        with self.web_session.get(media_url, headers=self._headers) as resp:
            if self.check_status(resp.status_code, media_url):
                response_text = resp.text
                response_text_as_dict = json.loads(response_text)
                return create_media_object(response_text_as_dict.get('media'))

    @check_expired_token
    def fetch_announcement(self, community_id: int, announcement_id: int) -> Optional[Announcement]:
        """Receive announcement object based on announcement id.

        :parameter community_id: The ID of the community the media belongs to.
        :parameter announcement_id: The ID of the announcement to fetch.
        :returns: :ref:`Announcement` or NoneType
        """
        announcement_url = self._api_communities_url + str(community_id) + "/notices/" + str(announcement_id)
        with self.web_session.get(announcement_url, headers=self._headers) as resp:
            if self.check_status(resp.status_code, announcement_url):
                response_text = resp.text
                response_text_as_dict = json.loads(response_text)
                return create_announcement_object(response_text_as_dict)

    def update_cache_from_notification(self) -> List[Notification]:
        """Grab a new post based from new notifications and add it to cache.

        Will also return the new notifications found.

        :returns: List[:class:`models.Notification`]
        """
        new_notifications = []
        try:
            notifications = self.get_user_notifications()
            if not notifications:
                return new_notifications

            new_notifications = self.get_new_notifications()
            for notification in new_notifications:
                self.__manage_notification_posts(notification)

        except Exception as e:
            if self.verbose:
                print(f"Failed to update Weverse Cache - {e}")
        return new_notifications

    def __manage_notification_posts(self, notification: Notification):
        """
        Manages the creation of Notification posts and comments.

        :param notification: Notification to create comments and posts for.
        """
        notification_type = self.determine_notification_type(notification.message)
        community = self.get_community_by_id(notification.community_id)
        if notification_type == 'comment':
            artist_comments = self.fetch_artist_comments(notification.community_id, notification.contents_id)
            comment = artist_comments[0]
            comment.post = self.get_post_by_id(comment.post_id)
            if comment.post:
                if comment.post.artist_comments:
                    comment.post.artist_comments.insert(0, comment)
                else:
                    comment.post.artist_comments = [comment]
            self.all_comments[comment.id] = comment

        elif notification_type in ["tofans", "post"]:
            post = self.create_post(community, notification.contents_id)
            if post:
                self.all_posts[post.id] = post
        elif notification_type == 'media':
            media = self.fetch_media(community.id, notification.contents_id)
            if media:
                self.all_media[media.id] = media
        elif notification_type == 'announcement':
            announcement = self.fetch_announcement(community.id, notification.contents_id)
            if announcement:
                self.all_announcements[announcement.id] = announcement

    def check_token_works(self):
        """
        Check if a token is invalid.

        :returns: (:class:`bool`) True if the token works.
        """
        with self.web_session.get(url=self._user_endpoint, headers=self._headers) as resp:
            self._expired_token = not resp.status_code == 200
            return not self._expired_token



from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from Weverse.models import Post


class Comment:
    r"""A Comment object that represents a Weverse Comment that belongs to an Artist.

    It is not suggested to create a Comment manually, but rather through the
    following method: :class:`Weverse.objects.create_comment_objects`

    The information retrieved on a Comment is directly from the Weverse API and altered to fit this class.

    .. container:: operations

        .. describe:: x == y

            Checks if two Comments have the same ID.

        .. describe:: x != y

            Checks if two Comments do not have the same ID.

        .. describe:: str(x)

            Returns the comment's body.

    Parameters
    ----------
    id: :class:`int`
        The ID of the comment.
    body: :class:`str`
        The comment content AKA the body of the message.
    comment_count: :class:`int`
        Amount of comments inside of this comment (replies).
    like_count: :class:`int`
        Amount of likes on the comment.
    has_my_like: :class:`bool`
        Whether the client has liked the comment.
    is_blind: :class:`bool`
        NOT SURE WHAT THIS IS
    post_id: :class:`int`
        The Post ID that the comment was created under.
    created_at:
        The time the comment was created.
    updated_at:
        The time the comment was updated.

    Attributes
    -----------
    id: int
        The ID of the comment.
    body: str
        The comment content AKA the body of the message.
    comment_count: int
        Amount of comments inside of this comment (replies).
    like_count: int
        Amount of likes on the comment.
    has_my_like: bool
        Whether the client has liked the comment.
    is_blind: bool
        NOT SURE WHAT THIS IS
    post_id: int
        The Post ID that the comment was created under.
    created_at:
        The time the comment was created.
    updated_at:
        The time the comment was updated.
    post: Post
        The Post Object the comment belongs to.
    """
    def __init__(self, **kwargs):
        self.id = kwargs.get('comment_id')
        self.body = kwargs.get('body')
        self.comment_count = kwargs.get('comment_count')
        self.like_count = kwargs.get('like_count')
        self.has_my_like = kwargs.get('has_my_like')
        self.is_blind = kwargs.get('is_blind')
        self.post_id = kwargs.get('post_id')
        self.created_at = kwargs.get('created_at')
        self.updated_at = kwargs.get('updated_at')
        self.post: Optional[Post] = None

    def __eq__(self, other):
        """Check if the IDs of the Comments are equal."""
        if not isinstance(other, Comment):
            return NotImplemented

        return self.id == other.id

    def __ne__(self, other):
        """Check if the IDs of the Comments are not equal."""
        return not self == other

    def __str__(self):
        """Returns the body of the comment."""
        return f"{self.body}"




from typing import Optional, TYPE_CHECKING
import aiohttp

if TYPE_CHECKING:
    from Weverse.models import Post


class Video:
    r"""A Video object that represents a Weverse Video that belongs to media or a post.

    It is not suggested to create a Video manually, but rather through the
    following method: :class:`Weverse.objects.create_video_objects`

    The information retrieved on a Video is directly from the Weverse API and altered to fit this class.

    Videos do not have unique IDs.

    .. container:: operations

        .. describe:: x == y

            Check if the Video URL and Post are the same.

        .. describe:: x != y

            Check if the Video objects are not equal.

        .. describe:: str(x)

            Returns the Video URL.

        .. describe:: len(x)

            Returns the length of the video in seconds.

    Parameters
    ----------
    video_url: int
        Direct URL to the video.
    thumbnail_url: str
        URL of the thumbnail.
    thumbnail_width: int
        Width of the thumbnail
    thumbnail_height: int
        Height of the thumbnail.
    length: int
        Duration of the video in seconds.

    Attributes
    -----------
    video_url: int
        Direct URL to the video.
    thumbnail_url: str
        URL of the thumbnail.
    thumbnail_width: int
        Width of the thumbnail
    thumbnail_height: int
        Height of the thumbnail.
    playtime: int
        Duration of the video in seconds.
    post: Optional[Post]
        The Post Object the video belongs to.
    """
    def __init__(self, **kwargs):
        self.video_url = kwargs.get('video_url')
        self.thumbnail_url = kwargs.get('thumbnail_url')
        self.thumbnail_width = kwargs.get('thumbnail_width')
        self.thumbnail_height = kwargs.get('thumbnail_height')
        self.length = kwargs.get('playtime')
        self.post: Optional[Post] = None
        self.community_id: Optional[int] = kwargs.get('community_id')

    def __eq__(self, other):
        """Check if the Video URL and Post are the same."""
        if not isinstance(other, Video):
            return NotImplemented

        video_url_check = self.video_url == other.video_url
        if other.post and self.post:
            return video_url_check and (self.post == other.post)
        return video_url_check

    def __ne__(self, other):
        """Check if the Video objects are not equal."""
        return not self == other

    def __str__(self):
        """Returns the Video URL."""
        return f"{self.video_url}"

    def __len__(self):
        """Returns the length of the video in seconds."""
        return self.length


class VideoStream(Video):
    """
    A video stream that inherits from :ref:`Video`

    Attributes
    ----------
    hls_path: str
        Link to the .m3u8 file.
    dash_path: str
        Link to the .mpd file.
    content_index: int
        Index of the videos.
    video_id: int
        Unique Video ID
    encoding_status: str
        Encoding Status of the Video Stream.
    type: str
        The type of video it is.
    video_width: int
        The width of the video
    video_height: int
        The height of the video
    is_vertical: bool
        If the video is vertical.
    caption_s3_paths: List[str]
        A list of caption paths.
    level: str
        The users that are able to view the video.
    base_url: str
        The base url of the video to access files.
    m3u8_urls: str
        Several urls for the resolution m3u8 files.
    """
    def __init__(self, **kwargs):
        super(VideoStream, self).__init__(**kwargs)
        self.hls_path = kwargs.get('hls_path')
        self.dash_path = kwargs.get('dash_path')
        self.content_index = kwargs.get("content_index")
        self.video_id = kwargs.get("video_id")
        self.encoding_status = kwargs.get("encoding_status")
        self.type = kwargs.get("type")
        self.video_width = kwargs.get("video_width")
        self.video_height = kwargs.get("video_height")
        self.is_vertical = kwargs.get("is_vertical")
        self.caption_s3_paths = kwargs.get("caption_s3_paths")
        self.level = kwargs.get("level")
        self.base_url = self.hls_path.replace("HLS.m3u8", "")
        self.m3u8_urls = [f"{self.base_url}HLS_{resolution}.m3u8" for resolution in [2560, 1440, 1080, 720, 540, 360]]



class Tab:
    r"""A Post object that represents a Weverse Post.

    It is not suggested to create a Post manually, but rather through the
    following method: :class:`Weverse.objects.create_post_objects`

    The information retrieved on a Post is directly from the Weverse API and altered to fit this class.

    .. container:: operations

        .. describe:: x == y

            Checks if two Tab objects have the same ID.

        .. describe:: x != y

            Check if the IDs of the Tab objects are not equal.

        .. describe:: str(x)

            Returns the Tab name.

    Parameters
    ----------
    tab_id: [Optional] int
        The ID of the Tab.
    name: [Optional] str
        The Tab name.

    Attributes
    -----------
    id: int
        The ID of the Tab.
    name: str
        The Tab name.

    """
    def __init__(self, tab_id=None, name=None):
        self.id = tab_id
        self.name = name

    def __eq__(self, other):
        """Check if the IDs of the Tab objects are equal."""
        if not isinstance(other, Tab):
            return NotImplemented

        return self.id == other.id

    def __ne__(self, other):
        """Check if the IDs of the Tab objects are not equal."""
        return not self == other

    def __str__(self):
        """Returns the Tab name."""
        return f"{self.name}"



from asyncio import iscoroutinefunction
from functools import wraps
from . import models

from .error import InvalidToken, PageNotFound, BeingRateLimited, LoginFailed, InvalidCredentials, NoHookFound
from .objects import create_tab_objects, create_community_objects, create_comment_objects, create_notification_objects,\
    create_media_object, create_post_objects, create_artist_objects, create_photo_objects, \
    iterate_community_media_categories, create_announcement_object, create_video_objects

__title__ = 'Weverse'
__author__ = 'MujyKun'
__license__ = 'MIT'
__version__ = '1.1.8.2'


def check_expired_token(func):
    """Decorator to reinstate a token if it is expired."""

    @wraps(func)
    async def wrap_async_function(self=None, *args, **kwargs):
        if self._expired_token:
            if self._refresh_token_exists:
                await self._refresh_token()
            else:
                if self._login_info_exists:
                    await self._try_login()
                    await self._wait_for_login()  # wait for login or an exception to occur.

        return await func(self, *args, **kwargs)

    @wraps(func)
    def wrap_sync_function(self=None, *args, **kwargs):
        if self._expired_token:
            if self._refresh_token_exists:
                self._refresh_token()
            else:
                if self._login_info_exists:
                    self._try_login()
        return func(self, *args, **kwargs)
    return wrap_sync_function if not iscoroutinefunction(func) else wrap_async_function


from .weverseclient import WeverseClient
from .weversesync import WeverseClientSync
from .weverseasync import WeverseClientAsync



from .comment import Comment
from .artist import Artist
from .community import Community
from .notification import Notification
from .photo import Photo
from .tab import Tab
from .video import Video, VideoStream
from .post import Post
from .media import Media
from .announcement import Announcement



from setuptools import setup, find_packages
import re

with open("README.md", "r") as f:
    long_description = f.read()

with open('Weverse/__init__.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1) or ''

if not version:
    raise RuntimeError('No Version Found.')


def parse_requirements(filename):
    """Load requirements from a pip requirements file."""
    line_iter = (line.strip() for line in open(filename))
    return [line for line in line_iter if line and not line.startswith("#")]


setup(
    name='Weverse',
    version=version,
    packages=find_packages(),
    install_requires=parse_requirements("requirements.txt"),
    url='https://github.com/MujyKun/Weverse/',
    license='MIT License',
    author='MujyKun',
    author_email='mujy@irenebot.com',
    description='Creates internal cache for the communities a user follows on weverse.io',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3.6',

)
